@@ -1,5 +1,8 @@
 #pragma once
 
+#include <unistd.h>
+#include <stdint.h>
+
 #include <osmocom/core/tdef.h>
 #include <osmocom/sigtran/osmo_ss7.h>
 #include "xua_msg.h"
@@ -16,8 +19,8 @@ int sua_addr_parse(struct osmo_sccp_addr *out, struct xua_msg *xua, uint16_t iei
 
 int sua_parse_gt(struct osmo_sccp_gt *gt, const uint8_t *data, unsigned int datalen);
 
-struct xua_msg *osmo_sccp_to_xua(struct msgb *msg);
-struct msgb *osmo_sua_to_sccp(struct xua_msg *xua);
+struct xua_msg *osmo_sccp_to_xua(const struct msgb *msg);
+struct msgb *osmo_sua_to_sccp(const struct xua_msg *xua);
 
 int sua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg);
 
@@ -28,11 +31,14 @@ void sua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsign
 void sua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 			   const uint32_t *aff_pc, unsigned int num_aff_pc, const uint32_t *ssn,
 			   const uint8_t cong_level, const char *info_string);
+void sua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+		     const uint32_t *aff_pc, unsigned int num_aff_pc, const uint32_t *aff_ssn,
+		     const uint32_t *smi, const char *info_string);
 void sua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 		 uint32_t dpc, uint16_t user, uint16_t cause, const char *info_str);
 
-struct osmo_mtp_prim *m3ua_to_xfer_ind(struct xua_msg *xua);
 int m3ua_hmdc_rx_from_l2(struct osmo_ss7_instance *inst, struct xua_msg *xua);
+struct msgb *m3ua_to_msg(struct xua_msg *xua);
 int m3ua_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua);
 void m3ua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 			   const uint32_t *aff_pc, unsigned int num_aff_pc,
@@ -41,9 +47,15 @@ void m3ua_tx_snm_congestion(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsi
 			    const uint32_t *aff_pc, unsigned int num_aff_pc,
 			    const uint32_t *concerned_dpc, const uint8_t *cong_level,
 			    const char *info_string);
+void m3ua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+		      const uint32_t *aff_pc, unsigned int num_aff_pc, const char *info_string);
 void m3ua_tx_dupu(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
 		  uint32_t dpc, uint16_t user, uint16_t cause, const char *info_str);
 
+void xua_tx_snm_available(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+			  const uint32_t *aff_pc, unsigned int num_aff_pc,
+			  const char *info_str, bool available);
+
 void xua_snm_rx_daud(struct osmo_ss7_asp *asp, struct xua_msg *xua);
 void xua_snm_rx_duna(struct osmo_ss7_asp *asp, struct osmo_ss7_as *as, struct xua_msg *xua);
 void xua_snm_rx_dava(struct osmo_ss7_asp *asp, struct osmo_ss7_as *as, struct xua_msg *xua);
@@ -54,11 +66,14 @@ int m3ua_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg);
 struct msgb *m3ua_msgb_alloc(const char *name);
 struct xua_msg *m3ua_xfer_from_data(const struct m3ua_data_hdr *data_hdr,
 				    const uint8_t *data, unsigned int data_len);
-struct m3ua_data_hdr *data_hdr_from_m3ua(struct xua_msg *xua);
+struct m3ua_data_hdr *data_hdr_from_m3ua(const struct xua_msg *xua);
 void m3ua_dh_to_xfer_param(struct osmo_mtp_transfer_param *param,
 			   const struct m3ua_data_hdr *mdh);
 void mtp_xfer_param_to_m3ua_dh(struct m3ua_data_hdr *mdh,
 				const struct osmo_mtp_transfer_param *param);
+int hmrt_mtp_xfer_request_l4_to_l3(struct osmo_ss7_instance *inst,
+				   const struct osmo_mtp_transfer_param *param,
+				   uint8_t *user_data, size_t user_data_len);
 
 
 extern const struct xua_msg_class m3ua_msg_class_mgmt;
@@ -71,6 +86,7 @@ extern const struct value_string m3ua_err_names[];
 extern const struct value_string m3ua_ntfy_type_names[];
 extern const struct value_string m3ua_ntfy_stchg_names[];
 extern const struct value_string m3ua_ntfy_other_names[];
+const char *m3ua_ntfy_info_name(uint16_t ntfy_status_type, uint16_t ntfy_status_info);
 
 struct xua_msg *m3ua_encode_notify(const struct osmo_xlm_prim_notify *npar);
 int m3ua_decode_notify(struct osmo_xlm_prim_notify *npar, void *ctx,
@@ -89,6 +105,9 @@ void xua_asp_send_xlm_prim_simple(struct osmo_ss7_asp *asp,
 void xua_snm_pc_available(struct osmo_ss7_as *as, const uint32_t *aff_pc,
                           unsigned int num_aff_pc, const char *info_str, bool available);
 
+void xua_tx_snm_daud(struct osmo_ss7_asp *asp, const uint32_t *rctx, unsigned int num_rctx,
+		     const uint32_t *aff_pc, unsigned int num_aff_pc, const char *info_str);
+
 enum ss7_asp_lm_timer {
 	/* 0 kept unused on purpose since it's handled specially by osmo_fsm */
 	SS7_ASP_LM_T_WAIT_ASP_UP = 1,
@@ -109,15 +128,13 @@ extern struct osmo_fsm xua_default_lm_fsm;
 extern const struct value_string m3ua_rkm_reg_status_vals[];
 extern const struct value_string m3ua_rkm_dereg_status_vals[];
 
-#define CS7_STR	"ITU-T Signaling System 7\n"
-#define PC_STR	"Point Code\n"
-#define INST_STR "An instance of the SS7 stack\n"
-
-int xua_as_transmit_msg(struct osmo_ss7_as *as, struct msgb *msg);
+int xua_as_transmit_msg(struct osmo_ss7_as *as, struct xua_msg *xua);
 
 int xua_find_as_for_asp(struct osmo_ss7_as **as, const struct osmo_ss7_asp *asp,
 			const struct xua_msg_part *rctx_ie);
 
+struct msgb *ipa_gen_ping(void);
+struct msgb *ipa_to_msg(struct xua_msg *xua);
 int ipa_tx_xua_as(struct osmo_ss7_as *as, struct xua_msg *xua);
 int ipa_rx_msg(struct osmo_ss7_asp *asp, struct msgb *msg, uint8_t sls);
 struct osmo_ss7_as *ipa_find_as_for_asp(struct osmo_ss7_asp *asp);
